import os
from typing import Optional

import rich_click as click

from truefoundry.common.auth_service_client import AuthServiceClient
from truefoundry.common.constants import TFY_API_KEY_ENV_KEY, TFY_HOST_ENV_KEY
from truefoundry.common.credential_file_manager import CredentialsFileManager
from truefoundry.common.credential_provider import EnvCredentialProvider
from truefoundry.common.entities import CredentialsFileContent, Token
from truefoundry.common.session import Session
from truefoundry.common.utils import relogin_error_message, resolve_tfy_host
from truefoundry.deploy.io.output_callback import OutputCallBack
from truefoundry.deploy.lib.const import (
    RICH_OUTPUT_CALLBACK,
)
from truefoundry.deploy.lib.messages import (
    PROMPT_ALREADY_LOGGED_OUT,
    PROMPT_LOGOUT_SUCCESSFUL,
)
from truefoundry.logger import logger


def login(
    api_key: Optional[str] = None,
    host: Optional[str] = None,
    relogin: bool = False,
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> bool:
    if TFY_API_KEY_ENV_KEY in os.environ and TFY_HOST_ENV_KEY in os.environ:
        logger.warning(
            "Skipping login because environment variables %s and "
            "%s are set and will be used when running truefoundry. "
            "If you want to relogin then unset these environment keys.",
            TFY_HOST_ENV_KEY,
            TFY_API_KEY_ENV_KEY,
        )
        return False

    if EnvCredentialProvider.can_provide():
        logger.warning(
            "TFY_API_KEY env var is already set. "
            "When running truefoundry, it will use the api key to authorize.\n"
            "Login will just save the credentials on disk."
        )

    host = resolve_tfy_host(host)

    with CredentialsFileManager() as cred_file:
        if not relogin and cred_file.exists():
            cred_file_content = cred_file.read()
            if host != cred_file_content.host:
                if click.confirm(
                    f"Already logged in to {cred_file_content.host!r}\n"
                    f"Do you want to relogin to {host!r}?"
                ):
                    return login(api_key=api_key, host=host, relogin=True)

            user_info = cred_file_content.to_token().to_user_info()
            output_hook.print_line(
                relogin_error_message(
                    f"Already logged in to {cred_file_content.host!r} as {user_info.user_id!r}",
                    host=host,
                )
            )
            return False

        if api_key:
            token = Token(access_token=api_key, refresh_token=None)
        else:
            auth_service = AuthServiceClient.from_tfy_host(tfy_host=host)
            # interactive login
            token = _login_with_device_code(
                base_url=host, auth_service=auth_service, output_hook=output_hook
            )

        cred_file_content = CredentialsFileContent(
            access_token=token.access_token,
            refresh_token=token.refresh_token,
            host=host,
        )
        cred_file.write(cred_file_content)

    user_info = token.to_user_info()
    output_hook.print_line(
        f"Successfully logged in to {cred_file_content.host!r} as {user_info.user_id!r}"
    )
    return True


def logout(
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> None:
    with CredentialsFileManager() as cred_file:
        if cred_file.delete():
            output_hook.print_line(PROMPT_LOGOUT_SUCCESSFUL)
        else:
            output_hook.print_line(PROMPT_ALREADY_LOGGED_OUT)


def get_access_token():
    # Get the access token from the session
    session = Session.new()
    return session.access_token


def _login_with_device_code(
    base_url: str,
    auth_service: AuthServiceClient,
    output_hook: OutputCallBack = RICH_OUTPUT_CALLBACK,
) -> Token:
    logger.debug("Logging in with device code")
    device_code = auth_service.get_device_code()
    auto_open_url = None
    message = "Please click on the above link if it is not automatically opened in a browser window."
    if device_code.complete_verification_url:
        auto_open_url = device_code.complete_verification_url
    elif device_code.verification_url:
        if device_code.message:
            message = device_code.message
        else:
            message = (
                f"Please open the following URL in a browser and enter the code {device_code.user_code} "
                f"when prompted: {device_code.verification_url}"
            )
    else:
        auto_open_url = device_code.get_user_clickable_url(auth_host=base_url)
    if auto_open_url:
        output_hook.print_line(f"Opening:- {auto_open_url}")
        click.launch(auto_open_url)
    output_hook.print_line(message)
    return auth_service.get_token_from_device_code(
        device_code=device_code.device_code,
        timeout=device_code.expires_in_seconds,
        poll_interval_seconds=device_code.interval_in_seconds,
    )



import truefoundry_sdk
from truefoundry_sdk.core.client_wrapper import SyncClientWrapper

from truefoundry.common.session import Session


def _get_base_url(self: SyncClientWrapper) -> str:
    if not isinstance(self._base_url, str):
        self._base_url = self._base_url()
    return self._base_url


class _LazyTrueFoundry(truefoundry_sdk.TrueFoundry):
    def _get_session(self) -> Session:
        if self._session is None:
            self._session = Session.new()
        return self._session

    def __init__(self):
        self._session = None
        # We patch this because SyncClientWrapper.get_base_url is expected to return a string but
        # we need to ensure that the base_url is only resolved once any requests are made
        # Flow:
        # internal http client tries to resolve base_url
        #   -> SyncClientWrapper.get_base_url (_get_base_url)
        #     -> self._get_session().tfy_host   # where self._get_session itself caches the session instance
        SyncClientWrapper.get_base_url = _get_base_url
        super().__init__(
            # since our base_url resolving is tied to session creation, we pass a callable that
            # the above _get_base_url will replace with the resolved result
            base_url=lambda: self._get_session().tfy_host,  # type: ignore
            api_key=lambda: self._get_session().access_token,
        )


client = _LazyTrueFoundry()



import re
from typing import Any, Dict, Optional

from truefoundry_sdk import ChatPromptManifest

_VAR_PATTERN = re.compile(r"\{\{(\w+)\}\}")


def render_text(template: str, variables: Dict[str, str]) -> str:
    return _VAR_PATTERN.sub(
        lambda m: (variables.get(m.group(1), f"{{{{{m.group(1)}}}}}")), template
    )


def render_prompt(
    prompt_template: ChatPromptManifest, variables: Optional[Dict[str, str]] = None
) -> Dict[str, Any]:
    """
    Renders a prompt template with the provided variables.
    Args:
        prompt_template (ChatPromptManifest): The prompt template to render.
        variables (Optional[Dict[str, str]]): A dictionary of variables to replace in the template.
            If None, the default variables from the prompt template will be used.

    Returns:
        Dict[str, Any]: A dictionary containing the rendered messages and model configuration.

    """
    # Render the messages in the prompt template with the provided variables and default variables
    rendered_messages = []
    variables = {**(prompt_template.variables or {}), **(variables or {})}
    messages = [message.copy(deep=True) for message in prompt_template.messages]

    # Render the messages with the provided variables, if any
    if variables:
        for msg in messages:
            if isinstance(msg.content, str):
                msg.content = render_text(template=msg.content, variables=variables)
            elif isinstance(msg.content, list):
                for part in msg.content:
                    if part.type == "image_url" and isinstance(part.image_url.url, str):
                        part.image_url.url = render_text(
                            part.image_url.url, variables=variables
                        )
                    elif part.type == "text" and isinstance(part.text, str):
                        part.text = render_text(part.text, variables=variables)

    rendered_messages = [message.dict() for message in messages]

    # Merge parameters from model_configuration and extra_parameters
    model_configuration = prompt_template.model_configuration
    if model_configuration:
        if model_configuration.provider:
            model = f"{model_configuration.provider}/{model_configuration.model}"
        else:
            model = model_configuration.model
        parameters = {
            **(
                model_configuration.parameters.dict()
                if model_configuration.parameters
                else {}
            ),
            **(model_configuration.extra_parameters or {}),
        }
    else:
        model = None
        parameters = {}

    return {
        "messages": rendered_messages,
        "model": model,
        "parameters": {k: v for k, v in parameters.items() if v is not None},
    }



"""Constants used by workflow tasks (e.g. Databricks OIDC and run-now)."""

from pathlib import Path

# Kubernetes default mount path for the service account token (OIDC).
SERVICE_ACCOUNT_TOKEN_PATH = Path("/var/run/secrets/kubernetes.io/serviceaccount/token")

# Databricks run-now API: idempotency_token max length; use only safe characters.
IDEMPOTENCY_TOKEN_MAX_LEN = 64
IDEMPOTENCY_TOKEN_ALLOWED_CHARS = set(
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-"
)



from __future__ import annotations

import os
from fnmatch import fnmatch
from typing import Any, Generator, List, Optional

import gitignorefile

from truefoundry.autodeploy.tools.base import (
    Event,
    Message,
    RequestEvent,
    ResponseEvent,
    Rule,
    Tool,
)
from truefoundry.pydantic_v1 import Field


class ListFiles(Tool):
    description = """
List all files.
If you want to find all json files recursively under directory a/b
the subdir should be a/b and pattern will be *.json

If you want to find all json files recuresively under current directory
the subdir should be . and pattern will be *.json
"""

    class Request(RequestEvent):
        sub_dir: str
        pattern: str = Field(..., description="Glob pattern. Avoid passing '*'")

    class Response(ResponseEvent):
        paths: List[str] = Field(
            ...,
            description="File paths under the given directory",
        )
        error: Optional[str] = Field(None)

    def __init__(self, project_root_path: str):
        self.project_root_path = project_root_path

    def run(self, request: ListFiles.Request) -> Generator[Event, Any, ResponseEvent]:
        yield Rule(message="[bold green]List Files[/]")
        yield Message(
            message=f"[bold cyan]Searching:[/] 🔍 Looking for files matching the pattern [magenta]{request.pattern}[/]"
        )

        paths: List[str] = []

        def gitignore(_):
            return False

        gitignore_path = os.path.join(self.project_root_path, ".gitignore")
        if os.path.exists(gitignore_path):
            gitignore = gitignorefile.parse(path=gitignore_path)

        path = os.path.join(self.project_root_path, request.sub_dir.strip(os.path.sep))
        if not os.path.exists(path):
            return ListFiles.Response(
                paths=None,
                error=f"Incorrect sub_dir {request.sub_dir}. Does not exist",
            )

        for root, dirs, ps in os.walk(
            path,
        ):
            root = root[len(path) :]
            if ".git" in dirs:
                dirs.remove(".git")
            paths.extend(
                os.path.join(root, p).lstrip(os.path.sep)
                for p in ps
                if fnmatch(p, request.pattern)
                and not gitignore(os.path.join(root, p).strip(os.path.sep))
            )
        if len(paths) > 0:
            yield Message(message=f"[bold green]Success:[/] Found {len(paths)} files.")
        else:
            yield Message(
                message=f"[red]Alert:[/] No files found matching the pattern {request.pattern}."
            )
        return ListFiles.Response(paths=paths)



# coding: utf-8

"""
TrueFoundry ML API

No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

The version of the OpenAPI document: 1.0.0
Generated by OpenAPI Generator (https://openapi-generator.tech)

Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations

import json
import pprint
import re  # noqa: F401
from typing import Optional

from truefoundry.ml._autogen.client.models.agent_open_api_tool_with_fqn import (
    AgentOpenAPIToolWithFQN,
)
from truefoundry.ml._autogen.client.models.agent_with_fqn import AgentWithFQN
from truefoundry.pydantic_v1 import (
    BaseModel,
    Field,
    StrictStr,
    conlist,
    constr,
    validator,
)


class AgentApp(BaseModel):
    """
    AgentApp
    """

    type: Optional[StrictStr] = "agent-app"
    tools: conlist(AgentOpenAPIToolWithFQN) = Field(
        default=..., description="Tools available to the Agent app"
    )
    agents: conlist(AgentWithFQN) = Field(
        default=..., description="Agents available to the Agent app"
    )
    root_agent: constr(strict=True, min_length=1) = Field(
        default=...,
        description="Root Agent for the app. This will be the first agent invoked",
    )
    __properties = ["type", "tools", "agents", "root_agent"]

    @validator("type")
    def type_validate_enum(cls, value):
        """Validates the enum"""
        if value is None:
            return value

        if value not in ("agent-app",):
            raise ValueError("must be one of enum values ('agent-app')")
        return value

    class Config:
        """Pydantic configuration"""

        allow_population_by_field_name = True
        validate_assignment = True

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.dict(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> AgentApp:
        """Create an instance of AgentApp from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self):
        """Returns the dictionary representation of the model using alias"""
        _dict = self.dict(by_alias=True, exclude={}, exclude_none=True)
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in tools (list)
        _items = []
        if self.tools:
            for _item in self.tools:
                if _item:
                    _items.append(_item.to_dict())
            _dict["tools"] = _items
        # override the default output from truefoundry.pydantic_v1 by calling `to_dict()` of each item in agents (list)
        _items = []
        if self.agents:
            for _item in self.agents:
                if _item:
                    _items.append(_item.to_dict())
            _dict["agents"] = _items
        return _dict

    @classmethod
    def from_dict(cls, obj: dict) -> AgentApp:
        """Create an instance of AgentApp from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return AgentApp.parse_obj(obj)

        _obj = AgentApp.parse_obj(
            {
                "type": obj.get("type") if obj.get("type") is not None else "agent-app",
                "tools": [
                    AgentOpenAPIToolWithFQN.from_dict(_item)
                    for _item in obj.get("tools")
                ]
                if obj.get("tools") is not None
                else None,
                "agents": [AgentWithFQN.from_dict(_item) for _item in obj.get("agents")]
                if obj.get("agents") is not None
                else None,
                "root_agent": obj.get("root_agent"),
            }
        )
        return _obj
